/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Philippe Wittenbergh <phiw@l-c-n.com>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Philippe Wittenbergh <phiw@l-c-n.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
/*
 * CSS rules to prevent author styling of pop-up menu <select> form controls.
 * Multi-line list-boxes (select[multiple], select[size]) are not affected.
 * Work-around for ugly display of pop-up menu widgets - bug 394063.
 */

select:not([size]):not([multiple]),
select[size="0"],
select[size="1"] {
  -moz-appearance: menulist !important;
  border-color: -moz-initial !important;
  background: -moz-initial !important;
  color: -moz-initial !important;
  padding: -moz-initial !important;
  font: -moz-list !important;
  border-width: -moz-initial !important;
  border-style: -moz-initial !important;
  text-indent: -moz-initial !important;
}

select:not([size]):not([multiple])[disabled],
select[size="0"][disabled],
select[size="1"][disabled] {
  color: GrayText !important;
}
