/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Vidur Apparao <vidur@netscape.com> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsISupports.idl"
#include "nsISVSchemaErrorHandler.idl"

/**
 * the original code came from mozilla/extensions/webservices/public/nsISchemaLoader.idl
 */

interface nsISVSchema;
interface nsISVSchemaType;
interface nsIDOMElement;
interface nsISVSchemaLoadListener;
interface nsIPrincipal;

[scriptable, uuid(ab374f43-73cf-4bba-925a-31ebdfc8d172)]
interface nsISVSchemaLoader : nsISupports {
  nsISVSchema load(in AString schemaURI);
  void loadAsync(in AString schemaURI,
                 in nsISVSchemaLoadListener listener);
  [noscript] void loadAsyncWithPrincipal(in AString schemaURI,
                                         in nsISVSchemaLoadListener listener,
                                         in nsIPrincipal principal);
  nsISVSchema processSchemaElement(in nsIDOMElement element,
                                   in nsISVSchemaErrorHandler aErrorHandler);
};

[scriptable, function, uuid(0fe151fe-4c7c-462c-9732-1b5cd211139b)]
interface nsISVSchemaLoadListener : nsISVSchemaErrorHandler {
  void onLoad(in nsISVSchema schema);
};

%{ C++
#define NS_SVSCHEMALOADER_CID                           \
{ /* 255be572-9d9f-40f1-8c92-d407f26c9330 */          \
 0x255be572, 0x9d9f, 0x40f1,                          \
{0x8c, 0x92, 0xd4, 0x07, 0xf2, 0x6c, 0x93, 0x30}}

#define NS_SVSCHEMALOADER_CONTRACTID "@mozilla.org/schemavalidator/schemaloader;1"

#define NS_SVBUILTINSCHEMACOLLECTION_CID                \
{ /* e62ddb80-f7c6-4c4b-96fa-2fda9e67eb4d */          \
 0xe62ddb80, 0xf7c6, 0x4c4b,                          \
{0x96, 0xfa, 0x2f, 0xda, 0x9e, 0x67, 0xeb, 0x4d}}

#define NS_SVBUILTINSCHEMACOLLECTION_CONTRACTID "@mozilla.org/schemavalidator/builtinschemacollection;1"

#define NS_ERROR_SCHEMA_NOT_SCHEMA_ELEMENT NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 1)
#define NS_ERROR_SCHEMA_UNKNOWN_TARGET_NAMESPACE NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 2)
#define NS_ERROR_SCHEMA_UNKNOWN_TYPE NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 3)
#define NS_ERROR_SCHEMA_UNKNOWN_PREFIX NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 4)
#define NS_ERROR_SCHEMA_INVALID_STRUCTURE NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 5)
#define NS_ERROR_SCHEMA_INVALID_TYPE_USAGE NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 6)
#define NS_ERROR_SCHEMA_MISSING_TYPE NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 7)
#define NS_ERROR_SCHEMA_FACET_VALUE_ERROR NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 8)
#define NS_ERROR_SCHEMA_LOADING_ERROR NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL, 9)
%}
