//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""afxres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#define _AFX_NO_SPLITTER_RESOURCES\r\n"
    "#define _AFX_NO_OLE_RESOURCES\r\n"
    "#define _AFX_NO_TRACKER_RESOURCES\r\n"
    "#define _AFX_NO_PROPERTY_RESOURCES\r\n"
    "\r\n"
    "#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)\r\n"
    "#ifdef _WIN32\r\n"
    "LANGUAGE 9, 1\r\n"
    "#pragma code_page(1252)\r\n"
    "#endif //_WIN32\r\n"
    "#include ""res\\testembed.rc2""  // non-Microsoft Visual C++ edited resources\r\n"
    "#include ""afxres.rc""         // Standard components\r\n"
    "#endif\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDR_MAINFRAME           ICON    DISCARDABLE     "res\\testembed.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDR_MAINFRAME           BITMAP  MOVEABLE PURE   "res\\Toolbar.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// Toolbar
//

IDR_MAINFRAME TOOLBAR DISCARDABLE  20, 18
BEGIN
    BUTTON      ID_NAV_BACK
    BUTTON      ID_NAV_FORWARD
    SEPARATOR
    BUTTON      ID_NAV_RELOAD
    BUTTON      ID_NAV_STOP
    SEPARATOR
    BUTTON      ID_NAV_HOME
    SEPARATOR
    BUTTON      ID_APP_ABOUT
END


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_MAINFRAME MENU PRELOAD DISCARDABLE 
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "&New Browser Window\tCtrl+N", ID_NEW_BROWSER
        MENUITEM SEPARATOR
        MENUITEM "&Open File...\tCtrl+O",       ID_FILE_OPEN
        MENUITEM "&Save Page As...\tCtrl+S",    ID_FILE_SAVE_AS
        MENUITEM SEPARATOR
        MENUITEM "&Print...\tCtrl-P",           ID_FILE_PRINT
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       ID_APP_EXIT
    END
    POPUP "&Edit"
    BEGIN
        MENUITEM "Cu&t\tCtrl+X",                ID_EDIT_CUT
        MENUITEM "&Copy\tCtrl+C",               ID_EDIT_COPY
        MENUITEM "&Paste\tCtrl+V",              ID_EDIT_PASTE
        MENUITEM SEPARATOR
        MENUITEM "Select &All",                 ID_EDIT_SELECT_ALL
        MENUITEM "Select &None",                ID_EDIT_SELECT_NONE
        MENUITEM SEPARATOR
        MENUITEM "&Find in This Page...\tCtrl+F", ID_EDIT_FIND
        MENUITEM SEPARATOR
        MENUITEM "Profiles...",                 ID_MANAGE_PROFILES
        MENUITEM SEPARATOR
        MENUITEM "Preferences...",              ID_EDIT_PREFERENCES
    END
    POPUP "&View"
    BEGIN
        MENUITEM "Page S&ource",                ID_VIEW_SOURCE
        MENUITEM "Page &Info",                  ID_VIEW_INFO
        MENUITEM SEPARATOR
        MENUITEM "&Toolbar",                    ID_VIEW_TOOLBAR
        MENUITEM "&Status Bar",                 ID_VIEW_STATUS_BAR
    END
    POPUP "&Go"
    BEGIN
        MENUITEM "&Back",                       ID_NAV_BACK
        MENUITEM "&Forward",                    ID_NAV_FORWARD
        MENUITEM "&Home",                       ID_NAV_HOME
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "&About TestEmbed...",         ID_APP_ABOUT
    END
    POPUP "&Tests"
    BEGIN
        MENUITEM "&Change URL",                 ID_TESTS_CHANGEURL
        MENUITEM "&Global &History",            ID_TESTS_GLOBALHISTORY
        MENUITEM "Create &File",                ID_TESTS_CREATEFILE
        MENUITEM "Create/Switch &Profile",      ID_TESTS_CREATEPROFILE
        MENUITEM "Add Tooltip Listener",        ID_TESTS_ADDTOOLTIPLISTENER
        MENUITEM "&Add Web Prog Listener",      ID_TESTS_ADDWEBPROGLISTENER
        MENUITEM SEPARATOR
        MENUITEM "Add History &Listener",       ID_TESTS_ADDHISTORYLISTENER
        MENUITEM "Remove History Listener",     ID_TESTS_REMOVEHISTORYLISTENER

        MENUITEM SEPARATOR
        POPUP "Add UriContentListener"
        BEGIN
            MENUITEM "Add from nsIWebBrowser",      ID_TESTS_ADDURICONTENTLISTENER_ADDFROMNSIWEBBROWSER

            MENUITEM "Add From nsIUriLoader",       ID_TESTS_ADDURICONTENTLISTENER_ADDFROMNSIURILOADER

            MENUITEM "OpenUri()",                   ID_TESTS_ADDURICONTENTLISTENER_OPENURI

        END
        MENUITEM "NSNewChannel/AyncOpen",       ID_TESTS_NSNEWCHANNEL
        MENUITEM "nsIIOService->NewURI()",      ID_TESTS_NSIIOSERVICENEWURI
        MENUITEM "nsIProtocolHandler->NewURI()", 
                                                ID_TESTS_NSIPROTOCOLHANDLERNEWURI

    END
    POPUP "&Interfaces"
    BEGIN
        MENUITEM "Run All Test Cases",          ID_INTERFACES_RUNALLTESTCASES

        MENUITEM SEPARATOR
        POPUP "nsI&File"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIFILE_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "InitWithPath",                ID_INTERFACES_NSIFILE_INITWITHPATH

            MENUITEM "AppendRelaticePath",          ID_INTERFACES_NSIFILE_APPENDRELATICEPATH

            MENUITEM "Exists",                      ID_INTERFACES_NSIFILE_EXISTS

            MENUITEM "Create",                      ID_INTERFACES_NSIFILE_CREATE

            MENUITEM "CopyTo",                      ID_INTERFACES_NSIFILE_COPYTO

            MENUITEM "MoveTo",                      ID_INTERFACES_NSIFILE_MOVETO

        END
        POPUP "nsI&SHistory"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSISHISTORY_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetCount",                    ID_INTERFACES_NSISHISTORY_GETCOUNT

            MENUITEM "GetIndex",                    ID_INTERFACES_NSISHISTORY_GETINDEX

            MENUITEM "GetMaxLength",                ID_INTERFACES_NSISHISTORY_GETMAXLENGTH

            MENUITEM "SetMaxLength",                ID_INTERFACES_NSISHISTORY_SETMAXLENGTH

            MENUITEM SEPARATOR
            MENUITEM "GetEntryAtIndex",             ID_INTERFACES_NSISHISTORY_GETENTRYATINDEX

            MENUITEM "PurgeHistory",                ID_INTERFACES_NSISHISTORY_PURGEHISTORY

            MENUITEM SEPARATOR
            MENUITEM "GetSHistoryEnumerator",       ID_INTERFACES_NSISHISTORY_GETSHISTORYENUMERATOR

            MENUITEM SEPARATOR
            POPUP "nsIHistoryEntry"
            BEGIN
                MENUITEM "Run All Tests",               ID_INTERFACES_NSISHISTORY_NSIHISTORYENTRY_RUNALLTESTS

                MENUITEM SEPARATOR
                MENUITEM "GetURI",                      ID_INTERFACES_NSISHISTORY_NSIHISTORYENTRY_GETURI

                MENUITEM "GetTitle",                    ID_INTERFACES_NSISHISTORY_NSIHISTORYENTRY_GETTITLE

                MENUITEM "GetIsSubFrame",               ID_INTERFACES_NSISHISTORY_NSIHISTORYENTRY_GETISSUBFRAME

            END
        END
        POPUP "nsI&WebNav"
        BEGIN
            MENUITEM "RunAllTests",                 ID_INTERFACES_NSIWEBNAV_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetCanGoBack",                ID_INTERFACES_NSIWEBNAV_GETCANGOBACK

            MENUITEM "GetCanGoForward",             ID_INTERFACES_NSIWEBNAV_GETCANGOFORWARD

            MENUITEM "GoBack",                      ID_INTERFACES_NSIWEBNAV_GOBACK

            MENUITEM "GoForward",                   ID_INTERFACES_NSIWEBNAV_GOFORWARD

            MENUITEM "GoToIndex",                   ID_INTERFACES_NSIWEBNAV_GOTOINDEX

            MENUITEM "LoadURI",                     ID_INTERFACES_NSIWEBNAV_LOADURI

            MENUITEM "Reload",                      ID_INTERFACES_NSIWEBNAV_RELOAD

            MENUITEM "Stop",                        ID_INTERFACES_NSIWEBNAV_STOP

            MENUITEM "GetDocument",                 ID_INTERFACES_NSIWEBNAV_GETDOCUMENT

            MENUITEM "GetCurrentURI",               ID_INTERFACES_NSIWEBNAV_GETCURRENTURI

            MENUITEM "GetReferringURI",             ID_INTERFACES_NSIWEBNAV_GETREFERRINGURI

            MENUITEM "GetSessionHistory",           ID_INTERFACES_NSIWEBNAV_GETSESSIONHISTORY

            MENUITEM "SetSessionHistory",           ID_INTERFACES_NSIWEBNAV_SETSESSIONHISTORY

        END
        POPUP "nsI&Channel"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSICHANNEL_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "SetOriginalURI",              ID_INTERFACES_NSICHANNEL_SETORIGINALURI

            MENUITEM "GetOriginalURI",              ID_INTERFACES_NSICHANNEL_GETORIGINALURI

            MENUITEM "GetURI",                      ID_INTERFACES_NSICHANNEL_GETURI

            MENUITEM "SetOwner",                    ID_INTERFACES_NSICHANNEL_SETOWNER

            MENUITEM "GetOwner",                    ID_INTERFACES_NSICHANNEL_GETOWNER

            MENUITEM "SetNotifications",            ID_INTERFACES_NSICHANNEL_SETNOTIFICATIONS

            MENUITEM "GetNotifications",            ID_INTERFACES_NSICHANNEL_GETNOTIFICATIONS

            MENUITEM "GetSecurityInfo",             ID_INTERFACES_NSICHANNEL_GETSECURITYINFO

            MENUITEM "SetContentType",              ID_INTERFACES_NSICHANNEL_SETCONTENTTYPE

            MENUITEM "GetContentType",              ID_INTERFACES_NSICHANNEL_GETCONTENTTYPE

            MENUITEM "SetContentCharset",           ID_INTERFACES_NSICHANNEL_SETCONTENTCHARSET

            MENUITEM "GetContentCharset",           ID_INTERFACES_NSICHANNEL_GETCONTENTCHARSET

            MENUITEM "SetContentLength",            ID_INTERFACES_NSICHANNEL_SETCONTENTLENGTH

            MENUITEM "GetContentLength",            ID_INTERFACES_NSICHANNEL_GETCONTENTLENGTH

            MENUITEM SEPARATOR
            MENUITEM "Open",                        ID_INTERFACES_NSICHANNEL_OPEN

            MENUITEM "AsyncOpen",                   ID_INTERFACES_NSICHANNEL_ASYNCOPEN

        END
        POPUP "nsI&Request"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIREQUEST_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetName",                     ID_INTERFACES_NSIREQUEST_GETNAME

            MENUITEM "IsPending",                   ID_INTERFACES_NSIREQUEST_ISPENDING

            MENUITEM "Cancel",                      ID_INTERFACES_NSIREQUEST_CANCEL

            MENUITEM "GetStatus",                   ID_INTERFACES_NSIREQUEST_GETSTATUS

            MENUITEM "Suspend",                     ID_INTERFACES_NSIREQUEST_SUSPEND

            MENUITEM "Resume",                      ID_INTERFACES_NSIREQUEST_RESUME

            MENUITEM "GetLoadGroup",                ID_INTERFACES_NSIREQUEST_GETLOADGROUP

            MENUITEM "SetLoadGroup",                ID_INTERFACES_NSIREQUEST_SETLOADGROUP

            MENUITEM "GetLoadFlags",                ID_INTERFACES_NSIREQUEST_GETLOADFLAGS

            MENUITEM "SetLoadFlags",                ID_INTERFACES_NSIREQUEST_SETLOADFLAGS

        END
        POPUP "nsIDirectoryService"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIDIRECTORYSERVICE_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "Init",                        ID_INTERFACES_NSIDIRECTORYSERVICE_INIT

            MENUITEM "Register Provider",           ID_INTERFACES_NSIDIRECTORYSERVICE_REGISTERPROVIDER

            MENUITEM "UnRegisterProvider",          ID_INTERFACES_NSIDIRECTORYSERVICE_UNREGISTERPROVIDER

        END
        POPUP "nsIDomWindow"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIDOMWINDOW_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetDomDocument",              ID_INTERFACES_NSIDOMWINDOW_GETDOMDOCUMENT

            MENUITEM "GetSelection",                ID_INTERFACES_NSIDOMWINDOW_GETSELECTION

            MENUITEM "GetParent",                   ID_INTERFACES_NSIDOMWINDOW_GETPARENT

            MENUITEM SEPARATOR
            MENUITEM "GetScrollBars",               ID_INTERFACES_NSIDOMWINDOW_GETSCROLLBARS

            MENUITEM "GetFrames",                   ID_INTERFACES_NSIDOMWINDOW_GETFRAMES

            MENUITEM "GetName",                     ID_INTERFACES_NSIDOMWINDOW_GETNAME

            MENUITEM SEPARATOR
            MENUITEM "GetTextZoom",                 ID_INTERFACES_NSIDOMWINDOW_GETTEXTZOOM

            MENUITEM "SetTextZoom",                 ID_INTERFACES_NSIDOMWINDOW_SETTEXTZOOM

            MENUITEM SEPARATOR
            MENUITEM "GetScsollX",                  ID_INTERFACES_NSIDOMWINDOW_GETSCSOLLX

            MENUITEM "GetScrollY",                  ID_INTERFACES_NSIDOMWINDOW_GETSCROLLY

            MENUITEM SEPARATOR
            MENUITEM "ScrollTo",                    ID_INTERFACES_NSIDOMWINDOW_SCROLLTO

            MENUITEM "ScrollBy",                    ID_INTERFACES_NSIDOMWINDOW_SCROLLBY

            MENUITEM "ScrollByLines",               ID_INTERFACES_NSIDOMWINDOW_SCROLLBYLINES

            MENUITEM "ScrollByPages",               ID_INTERFACES_NSIDOMWINDOW_SCROLLBYPAGES

            MENUITEM SEPARATOR
            MENUITEM "SizeToContent",               ID_INTERFACES_NSIDOMWINDOW_SIZETOCONTENT

        END
        POPUP "nsISelection"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSISELECTION_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetAnchorNode",               ID_INTERFACES_NSISELECTION_GETANCHORNODE

            MENUITEM "GetAnchorOffset",             ID_INTERFACES_NSISELECTION_GETANCHOROFFSET

            MENUITEM "GetFocusNode",                ID_INTERFACES_NSISELECTION_GETFOCUSNODE

            MENUITEM "GetFocusOffset",              ID_INTERFACES_NSISELECTION_GETFOCUSOFFSET

            MENUITEM "GetIsCollapsed",              ID_INTERFACES_NSISELECTION_GETISCOLLAPSED

            MENUITEM "GetRangeCount",               ID_INTERFACES_NSISELECTION_GETRANGECOUNT

            MENUITEM SEPARATOR
            MENUITEM "GetRangeAt",                  ID_INTERFACES_NSISELECTION_GETRANGEAT

            MENUITEM "Collapse",                    ID_INTERFACES_NSISELECTION_COLLAPSE

            MENUITEM "Extend",                      ID_INTERFACES_NSISELECTION_EXTEND

            MENUITEM "CollapseToStart",             ID_INTERFACES_NSISELECTION_COLLAPSETOSTART

            MENUITEM "CollapseToEnd",               ID_INTERFACES_NSISELECTION_COLLAPSETOEND

            MENUITEM SEPARATOR
            MENUITEM "ContainsNode",                ID_INTERFACES_NSISELECTION_CONTAINSNODE

            MENUITEM "SelectAllChildren",           ID_INTERFACES_NSISELECTION_SELECTALLCHILDREN

            MENUITEM SEPARATOR
            MENUITEM "AddRange",                    ID_INTERFACES_NSISELECTION_ADDRANGE

            MENUITEM "RemoveRange",                 ID_INTERFACES_NSISELECTION_REMOVERANGE

            MENUITEM "RemoveAllRanges",             ID_INTERFACES_NSISELECTION_REMOVEALLRANGES

            MENUITEM SEPARATOR
            MENUITEM "DeleteFromDocument",          ID_INTERFACES_NSISELECTION_DELETEFROMDOCUMENT

            MENUITEM "SelectionLanguageChange",     ID_INTERFACES_NSISELECTION_SELECTIONLANGUAGECHANGE

            MENUITEM "ToString",                    ID_INTERFACES_NSISELECTION_TOSTRING

        END
        POPUP "nsIProfile"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIPROFILE_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetProfileCount",             ID_INTERFACES_NSIPROFILE_GETPROFILECOUNT

            MENUITEM "GetCurrentProfile",           ID_INTERFACES_NSIPROFILE_GETCURRENTPROFILE

            MENUITEM "SetCurrentProfile",           ID_INTERFACES_NSIPROFILE_SETCURRENTPROFILE

            MENUITEM SEPARATOR
            MENUITEM "GetProfileList",              ID_INTERFACES_NSIPROFILE_GETPROFILELIST

            MENUITEM "ProfileExists",               ID_INTERFACES_NSIPROFILE_PROFILEEXISTS

            MENUITEM SEPARATOR
            MENUITEM "CreateNewProfile",            ID_INTERFACES_NSIPROFILE_CREATENEWPROFILE

            MENUITEM "RenameProfile",               ID_INTERFACES_NSIPROFILE_RENAMEPROFILE

            MENUITEM "DeleteProfile",               ID_INTERFACES_NSIPROFILE_DELETEPROFILE

            MENUITEM "CloneProfile",                ID_INTERFACES_NSIPROFILE_CLONEPROFILE

            MENUITEM "ShutDownCurrentProfile",      ID_INTERFACES_NSIPROFILE_SHUTDOWNCURRENTPROFILE

        END
        POPUP "nsIClipboardCommands"
        BEGIN
            MENUITEM "CanCopySelection",            ID_INTERFACES_NSICLIPBOARDCOMMANDS_CANCOPYSELECTION

            MENUITEM "CanCutSelection",             ID_INTERFACES_NSICLIPBOARDCOMMANDS_CANCUTSELECTION

            MENUITEM "CanPaste",                    ID_INTERFACES_NSICLIPBOARDCOMMANDS_CANPASTE

            MENUITEM SEPARATOR
            MENUITEM "CopySelection",               ID_INTERFACES_NSICLIPBOARDCOMMANDS_COPYSELECTION

            MENUITEM "CutSelection",                ID_INTERFACES_NSICLIPBOARDCOMMANDS_CUTSELECTION

            MENUITEM "Paste",                       ID_INTERFACES_NSICLIPBOARDCOMMANDS_PASTE

            MENUITEM "CopyLinkLocation",            ID_INTERFACES_NSICLIPBOARDCOMMANDS_COPYLINKLOCATION

            MENUITEM SEPARATOR
            MENUITEM "SelectAll",                   ID_INTERFACES_NSICLIPBOARDCOMMANDS_SELECTALL

            MENUITEM "SelectNone",                  ID_INTERFACES_NSICLIPBOARDCOMMANDS_SELECTNONE

        END
        POPUP "nsIObserverService"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIOBSERVERSERVICE_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "AddObservers",                ID_INTERFACES_NSIOBSERVERSERVICE_ADDOBSERVERS

            MENUITEM "EnumerateObservers",          ID_INTERFACES_NSIOBSERVERSERVICE_ENUMERATEOBSERVERS

            MENUITEM "NotifyObservers",             ID_INTERFACES_NSIOBSERVERSERVICE_NOTIFYOBSERVERS

            MENUITEM "RemoveObservers",             ID_INTERFACES_NSIOBSERVERSERVICE_REMOVEOBSERVERS

        END
        POPUP "nsIWebBrowser"
        BEGIN
            MENUITEM "",                            ID_INTERFACES_NSIWEBBROWSER

            MENUITEM "Run All Tests",               ID_INTERFACES_NSIWEBBROWSER_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "AddWebBrowserListener",       ID_INTERFACES_NSIWEBBROWSER_ADDWEBBROWSERLISTENER

            MENUITEM "RemoveWebBrowserListener",    ID_INTERFACES_NSIWEBBROWSER_REMOVEWEBBROWSERLISTENER

            MENUITEM "GetContainerWindow",          ID_INTERFACES_NSIWEBBROWSER_GETCONTAINERWINDOW

            MENUITEM "SetContainerWindow",          ID_INTERFACES_NSIWEBBROWSER_SETCONTAINERWINDOW

            MENUITEM "GetParentURIContentListener", ID_INTERFACES_NSIWEBBROWSER_GETPARENTURICONTENTLISTENER

            MENUITEM "SetParentURIContentListener", ID_INTERFACES_NSIWEBBROWSER_SETPARENTURICONTENTLISTENER

            MENUITEM "GetContentDOMWindow",         ID_INTERFACES_NSIWEBBROWSER_GETCONTENTDOMWINDOW

            MENUITEM SEPARATOR
            MENUITEM "nsIWBSetup->SetProperty",     ID_INTERFACES_NSIWEBBROWSER_NSIWBSETUPSETPROPERTY

        END
        POPUP "nsIWebProgress"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIWEBPROGRESS_RUNALLTESTS

            MENUITEM "AddProgressListener",         ID_INTERFACES_NSIWEBPROGRESS_ADDPROGRESSLISTENER

            MENUITEM "RemoveProgressListener",      ID_INTERFACES_NSIWEBPROGRESS_REMOVEPROGRESSLISTENER

            MENUITEM "GetDOMWindow",                ID_INTERFACES_NSIWEBPROGRESS_GETDOMWINDOW

            MENUITEM "IsLoadingDocument",           ID_INTERFACES_NSIWEBPROGRESS_ISLOADINGDOCUMENT

        END
        POPUP "nsIWebBrowserFind"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIWEBBROWSERFIND_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "SetSearchStringTest",         ID_INTERFACES_NSIWEBBROWSERFIND_SETSEARCHSTRINGTEST

            MENUITEM "GetSearchStringTest",         ID_INTERFACES_NSIWEBBROWSERFIND_GETSEARCHSTRINGTEST

            MENUITEM "FindNextTest",                ID_INTERFACES_NSIWEBBROWSERFIND_FINDNEXTTEST

            MENUITEM "SetFindBackwardsTest",        ID_INTERFACES_NSIWEBBROWSERFIND_SETFINDBACKWARDSTEST

            MENUITEM "GetFindBackwardsTest",        ID_INTERFACES_NSIWEBBROWSERFIND_GETFINDBACKWARDSTEST

            MENUITEM "SetWrapFindTest",             ID_INTERFACES_NSIWEBBROWSERFIND_SETWRAPFINDTEST

            MENUITEM "GetWrapFindTest",             ID_INTERFACES_NSIWEBBROWSERFIND_GETWRAPFINDTEST

            MENUITEM "SetEntireWordTest",           ID_INTERFACES_NSIWEBBROWSERFIND_SETENTIREWORDTEST

            MENUITEM "GetEntireWordTest",           ID_INTERFACES_NSIWEBBROWSERFIND_GETENTIREWORDTEST

            MENUITEM "SetMatchCase",                ID_INTERFACES_NSIWEBBROWSERFIND_SETMATCHCASE

            MENUITEM "GetMatchCase",                ID_INTERFACES_NSIWEBBROWSERFIND_GETMATCHCASE

            MENUITEM "SetSearchFrames",             ID_INTERFACES_NSIWEBBROWSERFIND_SETSEARCHFRAMES

            MENUITEM "GetSearchFrames",             ID_INTERFACES_NSIWEBBROWSERFIND_GETSEARCHFRAMES

        END
        POPUP "nsIEditingSession"
        BEGIN
            MENUITEM "RunAllTests",                 ID_INTERFACES_NSIEDITINGSESSION_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "Init",                        ID_INTERFACES_NSIEDITINGSESSION_INIT

            MENUITEM "MakeWindowEditable",          ID_INTERFACES_NSIEDITINGSESSION_MAKEWINDOWEDITABLE

            MENUITEM "WindowIsEditable",            ID_INTERFACES_NSIEDITINGSESSION_WINDOWISEDITABLE

            MENUITEM "GetEditorForWindow",          ID_INTERFACES_NSIEDITINGSESSION_GETEDITORFORWINDOW

            MENUITEM "SetupEditorOnWindow",         ID_INTERFACES_NSIEDITINGSESSION_SETUPEDITORONWINDOW

            MENUITEM "TearDownEditorOnWindow",      ID_INTERFACES_NSIEDITINGSESSION_TEARDOWNEDITORONWINDOW

        END
        POPUP "nsICommandManager"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSICOMMANDMANAGER_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "AddCommandObserver",          ID_INTERFACES_NSICOMMANDMANAGER_ADDCOMMANDOBSERVER

            MENUITEM "RemoveCommandObserver",       ID_INTERFACES_NSICOMMANDMANAGER_REMOVECOMMANDOBSERVER

            MENUITEM SEPARATOR
            MENUITEM "IsCommandeSupported",         ID_INTERFACES_NSICOMMANDMANAGER_ISCOMMANDESUPPORTED

            MENUITEM "IsCOmmandEnabled",            ID_INTERFACES_NSICOMMANDMANAGER_ISCOMMANDENABLED

            MENUITEM "GetCommandState",             ID_INTERFACES_NSICOMMANDMANAGER_GETCOMMANDSTATE

            MENUITEM "DoCOmmand",                   ID_INTERFACES_NSICOMMANDMANAGER_DOCOMMAND

        END
        POPUP "nsICommandParams"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSICOMMANDPARAMS_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "GetValueType",                ID_INTERFACES_NSICOMMANDPARAMS_GETVALUETYPE

            MENUITEM "GetBooleanValue",             ID_INTERFACES_NSICOMMANDPARAMS_GETBOOLEANVALUE

            MENUITEM "GetLongValue",                ID_INTERFACES_NSICOMMANDPARAMS_GETLONGVALUE

            MENUITEM "GetDoubleValue",              ID_INTERFACES_NSICOMMANDPARAMS_GETDOUBLEVALUE

            MENUITEM "GetStringValue",              ID_INTERFACES_NSICOMMANDPARAMS_GETSTRINGVALUE

            MENUITEM "GetCStringValue",             ID_INTERFACES_NSICOMMANDPARAMS_GETCSTRINGVALUE

            MENUITEM "GetISupportsValue",           ID_INTERFACES_NSICOMMANDPARAMS_GETISUPPORTSVALUE

            MENUITEM SEPARATOR
            MENUITEM "SetBooleanValue",             ID_INTERFACES_NSICOMMANDPARAMS_SETBOOLEANVALUE

            MENUITEM "SetLongValue",                ID_INTERFACES_NSICOMMANDPARAMS_SETLONGVALUE

            MENUITEM "SetDoubleValue",              ID_INTERFACES_NSICOMMANDPARAMS_SETDOUBLEVALUE

            MENUITEM "SetStringValue",              ID_INTERFACES_NSICOMMANDPARAMS_SETSTRINGVALUE

            MENUITEM "SetCStringValue",             ID_INTERFACES_NSICOMMANDPARAMS_SETCSTRINGVALUE

            MENUITEM "SetISupportsValue",           ID_INTERFACES_NSICOMMANDPARAMS_SETISUPPORTSVALUE

            MENUITEM SEPARATOR
            MENUITEM "RemoveValue",                 ID_INTERFACES_NSICOMMANDPARAMS_REMOVEVALUE

            MENUITEM "HasMoreElements",             ID_INTERFACES_NSICOMMANDPARAMS_HASMOREELEMENTS

            MENUITEM "First",                       ID_INTERFACES_NSICOMMANDPARAMS_FIRST

            MENUITEM "GetNext",                     ID_INTERFACES_NSICOMMANDPARAMS_GETNEXT

        END
        POPUP "nsIHttpChannel"
        BEGIN
            MENUITEM "Run All Tests",               ID_INTERFACES_NSIHTTPCHANNEL_RUNALLTESTS

            MENUITEM SEPARATOR
            MENUITEM "SetRequestMethod",            ID_INTERFACES_NSIHTTPCHANNEL_SETREQUESTMETHOD

            MENUITEM "GetRequestMethod",            ID_INTERFACES_NSIHTTPCHANNEL_GETREQUESTMETHOD

            MENUITEM "SetReferrer",                 ID_INTERFACES_NSIHTTPCHANNEL_SETREFERRER

            MENUITEM "GetReferrer",                 ID_INTERFACES_NSIHTTPCHANNEL_GETREFERRER

            MENUITEM "SetRequestHeader",            ID_INTERFACES_NSIHTTPCHANNEL_SETREQUESTHEADER

            MENUITEM "GetRequestHeader",            ID_INTERFACES_NSIHTTPCHANNEL_GETREQUESTHEADER

            MENUITEM "VisitRequestHeaders",         ID_INTERFACES_NSIHTTPCHANNEL_VISITREQUESTHEADERS

            MENUITEM "SetAllowPipelining",          ID_INTERFACES_NSIHTTPCHANNEL_SETALLOWPIPELINING

            MENUITEM "GetAllowPipelining",          ID_INTERFACES_NSIHTTPCHANNEL_GETALLOWPIPELINING

            MENUITEM "SetRedirectionLimit",         ID_INTERFACES_NSIHTTPCHANNEL_SETREDIRECTIONLIMIT

            MENUITEM "GetRedirectionLimit",         ID_INTERFACES_NSIHTTPCHANNEL_GETREDIRECTIONLIMIT

            MENUITEM SEPARATOR
            MENUITEM "GetResponseStatus",           ID_INTERFACES_NSIHTTPCHANNEL_GETRESPONSESTATUS

            MENUITEM "GetResponseStatusText",       ID_INTERFACES_NSIHTTPCHANNEL_GETRESPONSESTATUSTEXT

        END
    END
    POPUP "T&ools"
    BEGIN
        MENUITEM "Remove &GH Page",             ID_TOOLS_REMOVEGHPAGE
        MENUITEM "Remove &All GH",              ID_TOOLS_REMOVEALLGH
        MENUITEM "View &Logfile",               ID_TOOLS_VIEWLOGFILE
        MENUITEM "&Delete Logfile",             ID_TOOLS_DELETELOGFILE
        MENUITEM "Test &Your Method",           ID_TOOLS_TESTYOURMETHOD
        MENUITEM "Test Your &Method2",          ID_TOOLS_TESTYOURMETHOD2
    END
    POPUP "Ve&rify Bugs"
    BEGIN
        MENUITEM "&70228",                      ID_VERIFYBUGS_70228
        MENUITEM "90195",                       ID_VERIFYBUGS_90195
        MENUITEM "169617",                      ID_VERIFYBUGS_169617
        MENUITEM "170274",                      ID_VERIFYBUGS_170274
    END
END

IDR_CTXMENU_DOCUMENT MENU DISCARDABLE 
BEGIN
    POPUP "Document Context Menu"
    BEGIN
        MENUITEM "&Back",                       ID_NAV_BACK
        MENUITEM "&Forward",                    ID_NAV_FORWARD
        MENUITEM "&Reload",                     ID_NAV_RELOAD
        MENUITEM "&Stop",                       ID_NAV_STOP
        MENUITEM SEPARATOR
        MENUITEM "&View Page Source",           ID_VIEW_SOURCE
        MENUITEM "View Page &Info",             ID_VIEW_INFO
        MENUITEM SEPARATOR
        MENUITEM "Save &Page As...",            ID_FILE_SAVE_AS
        MENUITEM SEPARATOR
        MENUITEM "Select &All",                 ID_EDIT_SELECT_ALL
        MENUITEM "Select &None",                ID_EDIT_SELECT_NONE
        MENUITEM "&Copy",                       ID_EDIT_COPY
    END
END

IDR_CTXMENU_IMAGE MENU DISCARDABLE 
BEGIN
    POPUP "Image Context Menu"
    BEGIN
        MENUITEM "&Back",                       ID_NAV_BACK
        MENUITEM "&Forward",                    ID_NAV_FORWARD
        MENUITEM "&Reload",                     ID_NAV_RELOAD
        MENUITEM "&Stop",                       ID_NAV_STOP
        MENUITEM SEPARATOR
        MENUITEM "&View Page Source",           ID_VIEW_SOURCE
        MENUITEM "View &Image",                 ID_VIEW_IMAGE
        MENUITEM SEPARATOR
        MENUITEM "Save Image As...",            ID_SAVE_IMAGE_AS
        MENUITEM "Save &Page As...",            ID_FILE_SAVE_AS
    END
END

IDR_CTXMENU_LINK MENU DISCARDABLE 
BEGIN
    POPUP "Link Context Menu"
    BEGIN
        MENUITEM "Open Link In New Window",     ID_OPEN_LINK_IN_NEW_WINDOW
        MENUITEM SEPARATOR
        MENUITEM "&Back",                       ID_NAV_BACK
        MENUITEM "&Forward",                    ID_NAV_FORWARD
        MENUITEM "&Reload",                     ID_NAV_RELOAD
        MENUITEM "&Stop",                       ID_NAV_STOP
        MENUITEM SEPARATOR
        MENUITEM "&View Page Source",           ID_VIEW_SOURCE
        MENUITEM SEPARATOR
        MENUITEM "Save &Page As...",            ID_FILE_SAVE_AS
        MENUITEM "Save &Link As...",            ID_SAVE_LINK_AS
        MENUITEM "&Copy Link Location",         ID_COPY_LINK_LOCATION
    END
END

IDR_CTXMENU_TEXT MENU DISCARDABLE 
BEGIN
    POPUP "Selection Context Menu"
    BEGIN
        MENUITEM "Cu&t",                        ID_EDIT_CUT
        MENUITEM "&Copy",                       ID_EDIT_COPY
        MENUITEM "&Paste",                      ID_EDIT_PASTE
        MENUITEM "Select &All",                 ID_EDIT_SELECT_ALL
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDR_MAINFRAME ACCELERATORS PRELOAD MOVEABLE PURE 
BEGIN
    "C",            ID_EDIT_COPY,           VIRTKEY, CONTROL, NOINVERT
    "F",            ID_EDIT_FIND,           VIRTKEY, CONTROL, NOINVERT
    "N",            ID_NEW_BROWSER,         VIRTKEY, CONTROL, NOINVERT
    "O",            ID_FILE_OPEN,           VIRTKEY, CONTROL, NOINVERT
    "S",            ID_FILE_SAVE_AS,        VIRTKEY, CONTROL, NOINVERT
    "V",            ID_EDIT_PASTE,          VIRTKEY, CONTROL, NOINVERT
    VK_BACK,        ID_EDIT_UNDO,           VIRTKEY, ALT, NOINVERT
    VK_DELETE,      ID_EDIT_CUT,            VIRTKEY, SHIFT, NOINVERT
    VK_F6,          ID_NEXT_PANE,           VIRTKEY, NOINVERT
    VK_F6,          ID_PREV_PANE,           VIRTKEY, SHIFT, NOINVERT
    VK_INSERT,      ID_EDIT_COPY,           VIRTKEY, CONTROL, NOINVERT
    VK_INSERT,      ID_EDIT_PASTE,          VIRTKEY, SHIFT, NOINVERT
    "X",            ID_EDIT_CUT,            VIRTKEY, CONTROL, NOINVERT
    "Z",            ID_EDIT_UNDO,           VIRTKEY, CONTROL, NOINVERT
    VK_ESCAPE,      ID_NAV_STOP,            VIRTKEY 
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOG DISCARDABLE  0, 0, 235, 55
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About TestEmbed"
FONT 8, "MS Sans Serif"
BEGIN
    ICON            IDR_MAINFRAME,IDC_STATIC,11,17,20,20
    LTEXT           "Mozilla TestEmbed Version 1.0",IDC_STATIC,40,10,119,8,
                    SS_NOPREFIX
    LTEXT           "Copyright (C) 2001",IDC_STATIC,40,25,119,8
    DEFPUSHBUTTON   "OK",IDOK,178,7,50,14,WS_GROUP
END

IDD_PROMPT_DIALOG DIALOG DISCARDABLE  0, 0, 249, 92
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Prompt Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,130,69,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,189,69,50,14
    LTEXT           "Please enter something....",IDC_PROMPT_TEXT,10,7,229,29
    EDITTEXT        IDC_PROMPT_ANSWER,7,39,232,13,ES_AUTOHSCROLL
    CONTROL         "Save Answer",IDC_CHECK_SAVE_PASSWORD,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,57,229,10
END

IDD_PROMPT_PASSWORD_DIALOG DIALOG DISCARDABLE  0, 0, 248, 93
STYLE DS_SYSMODAL | DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | 
    WS_SYSMENU
CAPTION "Password Entry Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Please enter your password:",IDC_PROMPT_TEXT,5,7,232,20
    DEFPUSHBUTTON   "OK",IDOK,129,75,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,188,75,50,14
    EDITTEXT        IDC_PASSWORD,5,37,233,12,ES_PASSWORD | ES_AUTOHSCROLL
    CONTROL         "Save Password",IDC_CHECK_SAVE_PASSWORD,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,57,229,10
END

IDD_PROMPT_USERPASS_DIALOG DIALOG DISCARDABLE  0, 0, 214, 123
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Username and Password Required"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "This operation requires authentication.  Please enter your user name and password:",
                    IDC_PROMPT_TEXT,6,9,171,25
    LTEXT           "User Name:",IDC_USERNAME_LABEL,6,39,39,9
    EDITTEXT        IDC_USERNAME,49,39,153,12,ES_AUTOHSCROLL
    LTEXT           "Password:",IDC_PASSWORD_LABEL,6,58,36,8
    EDITTEXT        IDC_PASSWORD,49,56,153,12,ES_PASSWORD | ES_AUTOHSCROLL
    CONTROL         "Save User Name and Password",IDC_CHECK_SAVE_PASSWORD,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,6,78,201,10
    DEFPUSHBUTTON   "OK",IDOK,40,101,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,95,101,50,14
END

IDD_PROFILES DIALOG DISCARDABLE  0, 0, 196, 95
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Manage Profiles"
FONT 8, "MS Sans Serif"
BEGIN
    LISTBOX         IDC_LIST1,79,17,110,44,LBS_NOINTEGRALHEIGHT | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "Rename...",IDC_PROF_RENAME,7,7,50,14
    PUSHBUTTON      "Delete",IDC_PROF_DELETE,7,27,50,14
    PUSHBUTTON      "New...",IDC_PROF_NEW,7,48,50,14
    DEFPUSHBUTTON   "OK",IDOK,139,74,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,80,74,50,14
    CONTROL         "Ask At Startup",IDC_CHECK_ASK_AT_START,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,75,61,10
END

IDD_PROFILE_NEW DIALOG DISCARDABLE  0, 0, 177, 79
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "New Profile"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_NEW_PROF_NAME,39,16,122,14,ES_AUTOHSCROLL
    COMBOBOX        IDC_LOCALE_COMBO,39,35,67,30,CBS_DROPDOWNLIST | CBS_SORT | 
                    WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    DEFPUSHBUTTON   "OK",IDOK,120,58,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,63,58,50,14
    RTEXT           "Name:",IDC_STATIC,13,18,22,8
    RTEXT           "Locale:",IDC_STATIC,11,37,24,8
END

IDD_PROFILE_RENAME DIALOG DISCARDABLE  0, 0, 177, 76
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Rename Profile"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_NEW_NAME,50,16,113,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,120,55,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,63,55,50,14
    RTEXT           "New Name:",IDC_STATIC,7,18,38,8
END

IDD_FINDDLG DIALOG DISCARDABLE  30, 73, 242, 87
STYLE DS_MODALFRAME | DS_3DLOOK | DS_CONTEXTHELP | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
CAPTION "Find"
FONT 8, "MS Shell Dlg"
BEGIN
    LTEXT           "Fi&nd what:",-1,4,8,42,8
    EDITTEXT        IDC_FIND_EDIT,47,7,128,12,ES_AUTOHSCROLL | WS_GROUP
    CONTROL         "Wra&p around",IDC_WRAP_AROUND,"Button",BS_AUTOCHECKBOX | 
                    WS_GROUP | WS_TABSTOP,47,44,65,12
    CONTROL         "Match &case",IDC_MATCH_CASE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,47,26,64,12
    CONTROL         "Search &backwards",IDC_SEARCH_BACKWARDS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,47,62,75,12
    DEFPUSHBUTTON   "&Find Next",IDOK,188,5,50,14,WS_GROUP
    PUSHBUTTON      "Cancel",IDCANCEL,188,23,50,14
END

IDD_PRINT_PROGRESS_DIALOG DIALOG DISCARDABLE  0, 0, 294, 55
STYLE DS_SYSMODAL | DS_MODALFRAME | DS_CENTER | WS_POPUP | WS_VISIBLE | 
    WS_CAPTION
CAPTION "Printing..."
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Cancel",IDCANCEL,122,34,50,14
    LTEXT           "Document:",IDC_PPD_DOC_TITLE_STATIC,7,7,36,8
    LTEXT           "",IDC_PPD_DOC_TXT,43,7,244,8
END

IDD_PREFS_START_PAGE DIALOG DISCARDABLE  0, 0, 284, 150
STYLE WS_CHILD | WS_DISABLED | WS_CAPTION
CAPTION "Startup"
FONT 8, "MS Sans Serif"
BEGIN
    GROUPBOX        "TestEmbed starts with",IDC_STATIC,6,10,271,49
    CONTROL         "Bla&nk Page",IDC_RADIO_BLANK_PAGE,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,16,24,53,10
    CONTROL         "&Home Page",IDC_RADIO_HOME_PAGE,"Button",
                    BS_AUTORADIOBUTTON,16,40,53,10
    GROUPBOX        "Home Page",IDC_STATIC,6,67,271,51
    LTEXT           "Clicking the Home button will take you to this page",
                    IDC_STATIC,17,80,234,8
    LTEXT           "Loc&ation:",IDC_STATIC,17,98,31,8
    EDITTEXT        IDC_EDIT_HOMEPAGE,53,95,211,14,ES_AUTOHSCROLL
END

IDD_URLDIALOG DIALOG DISCARDABLE  0, 0, 209, 130
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "URL Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,131,7,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,131,22,50,14
    LTEXT           "Enter URL here:",IDC_ENTER_URLTEXT,64,55,71,8
    EDITTEXT        IDC_URLFIELD,25,68,134,14,ES_AUTOHSCROLL
    COMBOBOX        IDC_COMBO1,25,29,86,72,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    CONTROL         "Select Load Flag:",IDC_CHKURLFLAG,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,32,14,80,10
    LTEXT           "Protocol (optional):",IDC_STATIC,51,97,60,8
    COMBOBOX        IDC_COMBO2,58,109,48,66,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
END

IDD_RUNTESTSDLG DIALOG DISCARDABLE  0, 0, 331, 186
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Test Results"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Close",IDOK,274,165,50,14
    CONTROL         "List1",IDC_LIST1,"SysListView32",LVS_REPORT | 
                    LVS_NOSORTHEADER | WS_BORDER | WS_TABSTOP,7,7,317,154
END

IDD_QAFINDDLG DIALOG DISCARDABLE  0, 0, 186, 95
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Text Search Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,129,7,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,129,24,50,14
    EDITTEXT        IDC_TEXTFIELD,53,62,71,14,ES_AUTOHSCROLL
    LTEXT           "Enter text here:",IDC_STATIC,62,46,49,8
END

IDD_WEBPROGDLG DIALOG DISCARDABLE  0, 0, 186, 95
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Web Prog Lstnr Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,129,7,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,129,24,50,14
    LTEXT           "Web Prog Notification Flag",IDC_STATIC,29,46,97,8
    COMBOBOX        IDC_WPCOMBO,22,65,105,70,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
END


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,0,0,1
 PRODUCTVERSION 1,0,0,1
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904B0"
        BEGIN
            VALUE "CompanyName", "\0"
            VALUE "FileDescription", "TestEmbed MFC Application\0"
            VALUE "FileVersion", "1, 0, 0, 1\0"
            VALUE "InternalName", "TestEmbed\0"
            VALUE "LegalCopyright", "Copyright (C) 2001\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "OriginalFilename", "TestEmbed.EXE\0"
            VALUE "ProductName", "TestEmbed Application\0"
            VALUE "ProductVersion", "1, 0, 0, 1\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // !_MAC


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 228
        TOPMARGIN, 7
        BOTTOMMARGIN, 48
    END

    IDD_PROFILES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 189
        TOPMARGIN, 7
        BOTTOMMARGIN, 88
    END

    IDD_PROFILE_NEW, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 170
        TOPMARGIN, 7
        BOTTOMMARGIN, 72
    END

    IDD_PROFILE_RENAME, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 170
        TOPMARGIN, 7
        BOTTOMMARGIN, 69
    END

    IDD_FINDDLG, DIALOG
    BEGIN
        RIGHTMARGIN, 238
        BOTTOMMARGIN, 62
    END

    IDD_PRINT_PROGRESS_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 287
        TOPMARGIN, 7
        BOTTOMMARGIN, 48
    END

    IDD_URLDIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 186
        TOPMARGIN, 7
        BOTTOMMARGIN, 123
    END

    IDD_RUNTESTSDLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 324
        TOPMARGIN, 7
        BOTTOMMARGIN, 179
    END

    IDD_QAFINDDLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 88
    END

    IDD_WEBPROGDLG, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 88
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog Info
//

IDD_URLDIALOG DLGINIT
BEGIN
    IDC_COMBO1, 0x403, 6, 0
0x4f4e, 0x454e, 0x0020, 
    IDC_COMBO1, 0x403, 5, 0
0x414d, 0x4b53, "\000" 
    IDC_COMBO1, 0x403, 8, 0
0x5349, 0x4c5f, 0x4e49, 0x004b, 
    IDC_COMBO1, 0x403, 15, 0
0x5942, 0x4150, 0x5353, 0x485f, 0x5349, 0x4f54, 0x5952, "\000" 
    IDC_COMBO1, 0x403, 16, 0
0x4552, 0x4c50, 0x4341, 0x5f45, 0x4948, 0x5453, 0x524f, 0x0059, 
    IDC_COMBO1, 0x403, 13, 0
0x5942, 0x4150, 0x5353, 0x435f, 0x4341, 0x4548, "\000" 
    IDC_COMBO1, 0x403, 13, 0
0x5942, 0x4150, 0x5353, 0x505f, 0x4f52, 0x5958, "\000" 
    IDC_COMBO1, 0x403, 15, 0
0x4843, 0x5241, 0x4553, 0x5f54, 0x4843, 0x4e41, 0x4547, "\000" 
    IDC_COMBO1, 0x403, 8, 0
0x4552, 0x5246, 0x5345, 0x0048, 
    IDC_COMBO2, 0x403, 6, 0
0x7468, 0x7074, 0x0020, 
    IDC_COMBO2, 0x403, 6, 0
0x7468, 0x7074, 0x0073, 
    IDC_COMBO2, 0x403, 4, 0
0x7466, 0x0070, 
    IDC_COMBO2, 0x403, 5, 0
0x6966, 0x656c, "\000" 
    IDC_COMBO2, 0x403, 11, 0
0x616a, 0x6176, 0x6373, 0x6972, 0x7470, "\000" 
    IDC_COMBO2, 0x403, 6, 0
0x6261, 0x756f, 0x0074, 
    IDC_COMBO2, 0x403, 5, 0
0x6164, 0x6174, "\000" 
    IDC_COMBO2, 0x403, 7, 0
0x616d, 0x6c69, 0x6f74, "\000" 
    IDC_COMBO2, 0x403, 5, 0
0x6d69, 0x7061, "\000" 
    0
END

IDD_WEBPROGDLG DLGINIT
BEGIN
    IDC_WPCOMBO, 0x403, 21, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x4554, 0x525f, 0x5145, 0x4555, 
0x5453, "\000" 
    IDC_WPCOMBO, 0x403, 22, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x4554, 0x445f, 0x434f, 0x4d55, 
0x4e45, 0x0054, 
    IDC_WPCOMBO, 0x403, 21, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x4554, 0x4e5f, 0x5445, 0x4f57, 
0x4b52, "\000" 
    IDC_WPCOMBO, 0x403, 20, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x4554, 0x575f, 0x4e49, 0x4f44, 
0x0057, 
    IDC_WPCOMBO, 0x403, 17, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x4554, 0x415f, 0x4c4c, "\000" 
    IDC_WPCOMBO, 0x403, 16, 0
0x4f4e, 0x4954, 0x5946, 0x505f, 0x4f52, 0x5247, 0x5345, 0x0053, 
    IDC_WPCOMBO, 0x403, 14, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4154, 0x5554, 0x0053, 
    IDC_WPCOMBO, 0x403, 16, 0
0x4f4e, 0x4954, 0x5946, 0x535f, 0x4345, 0x5255, 0x5449, 0x0059, 
    IDC_WPCOMBO, 0x403, 16, 0
0x4f4e, 0x4954, 0x5946, 0x4c5f, 0x434f, 0x5441, 0x4f49, 0x004e, 
    IDC_WPCOMBO, 0x403, 11, 0
0x4f4e, 0x4954, 0x5946, 0x415f, 0x4c4c, "\000" 
    0
END


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE PRELOAD DISCARDABLE 
BEGIN
    IDR_MAINFRAME           "testembed"
END

STRINGTABLE PRELOAD DISCARDABLE 
BEGIN
    AFX_IDS_APP_TITLE       "testembed"
    AFX_IDS_IDLEMESSAGE     "Ready"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_APP_ABOUT            "Display program information, version number and copyright\nAbout"
    ID_APP_EXIT             "Quit the application; prompts to save documents\nExit"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_NEXT_PANE            "Switch to the next window pane\nNext Pane"
    ID_PREV_PANE            "Switch back to the previous window pane\nPrevious Pane"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_WINDOW_SPLIT         "Split the active window into panes\nSplit"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_EDIT_CLEAR           "Erase the selection\nErase"
    ID_EDIT_CLEAR_ALL       "Erase everything\nErase All"
    ID_EDIT_COPY            "Copy the selection and put it on the Clipboard\nCopy"
    ID_EDIT_CUT             "Cut the selection and put it on the Clipboard\nCut"
    ID_EDIT_FIND            "Find the specified text\nFind"
    ID_EDIT_PASTE           "Insert Clipboard contents\nPaste"
    ID_EDIT_REPEAT          "Repeat the last action\nRepeat"
    ID_EDIT_REPLACE         "Replace specific text with different text\nReplace"
    ID_EDIT_SELECT_ALL      "Select the entire document\nSelect All"
    ID_EDIT_UNDO            "Undo the last action\nUndo"
    ID_EDIT_REDO            "Redo the previously undone action\nRedo"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_VIEW_TOOLBAR         "Show or hide the toolbar\nToggle ToolBar"
    ID_VIEW_STATUS_BAR      "Show or hide the status bar\nToggle StatusBar"
END

STRINGTABLE DISCARDABLE 
BEGIN
    AFX_IDS_SCSIZE          "Change the window size"
    AFX_IDS_SCMOVE          "Change the window position"
    AFX_IDS_SCMINIMIZE      "Reduce the window to an icon"
    AFX_IDS_SCMAXIMIZE      "Enlarge the window to full size"
    AFX_IDS_SCNEXTWINDOW    "Switch to the next document window"
    AFX_IDS_SCPREVWINDOW    "Switch to the previous document window"
    AFX_IDS_SCCLOSE         "Close the active window and prompts to save the documents"
END

STRINGTABLE DISCARDABLE 
BEGIN
    AFX_IDS_SCRESTORE       "Restore the window to normal size"
    AFX_IDS_SCTASKLIST      "Activate Task List"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_NAV_RELOAD           "Tool Tip\nTool Tip"
    ID_EDIT_SELECT_NONE     "Select nothing in the document\nSelect None"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_INTERFACES_TEMP      "Temporary"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#define _AFX_NO_SPLITTER_RESOURCES
#define _AFX_NO_OLE_RESOURCES
#define _AFX_NO_TRACKER_RESOURCES
#define _AFX_NO_PROPERTY_RESOURCES

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE 9, 1
#pragma code_page(1252)
#endif //_WIN32
#include "res\testembed.rc2"  // non-Microsoft Visual C++ edited resources
#include "afxres.rc"         // Standard components
#endif

/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

